# Owner(s): ["module: fx"]

from __future__ import annotations  # type: ignore[attr-defined]

import torch
from torch.fx import symbolic_trace


class A:
    def __call__(self, x: torch.Tensor):
        return torch.add(x, x)


# No forward references
class M1(torch.nn.Module):
    def forward(self, x: torch.Tensor, a: A) -> torch.Tensor:
        return a(x)


# Forward references
class M2(torch.nn.Module):
    def forward(self, x: torch.Tensor, a: A) -> torch.Tensor:
        return a(x)


# Non-torch annotation with no internal forward references
class M3(torch.nn.Module):
    def forward(self, x: list[torch.Tensor], a: A) -> torch.Tensor:
        return a(x[0])


# Non-torch annotation with internal forward references
class M4(torch.nn.Module):
    def forward(self, x: list[torch.Tensor], a: A) -> torch.Tensor:
        return a(x[0])


x = torch.rand(2, 3)

ref = torch.add(x, x)

traced1 = symbolic_trace(M1())
res1 = traced1(x, A())
if not torch.all(torch.eq(ref, res1)):
    raise AssertionError("res1 does not match ref")

traced2 = symbolic_trace(M2())
res2 = traced2(x, A())
if not torch.all(torch.eq(ref, res2)):
    raise AssertionError("res2 does not match ref")

traced3 = symbolic_trace(M3())
res3 = traced3([x], A())
if not torch.all(torch.eq(ref, res3)):
    raise AssertionError("res3 does not match ref")

traced4 = symbolic_trace(M4())
res4 = traced4([x], A())
if not torch.all(torch.eq(ref, res4)):
    raise AssertionError("res4 does not match ref")
