# Defined in torch/csrc/export/pt2_archive_constants.h

ARCHIVE_ROOT_NAME: str = ...
ARCHIVE_FORMAT_PATH: str = ...
ARCHIVE_FORMAT_VALUE: str = ...
ARCHIVE_VERSION_PATH: str = ...
ARCHIVE_VERSION_VALUE: str = ...
MODELS_DIR: str = ...
MODELS_FILENAME_FORMAT: str = ...
AOTINDUCTOR_DIR: str = ...
MTIA_DIR: str = ...
WEIGHTS_DIR: str = ...
WEIGHTS_CONFIG_FILENAME_FORMAT: str = ...
WEIGHT_FILENAME_PREFIX: str = ...
CONSTANTS_DIR: str = ...
CONSTANTS_CONFIG_FILENAME_FORMAT: str = ...
TENSOR_CONSTANT_FILENAME_PREFIX: str = ...
CUSTOM_OBJ_FILENAME_PREFIX: str = ...
SAMPLE_INPUTS_DIR: str = ...
SAMPLE_INPUTS_FILENAME_FORMAT: str = ...
EXECUTORCH_DIR: str = ...
EXTRA_DIR: str = ...
MODULE_INFO_PATH: str = ...
XL_MODEL_WEIGHTS_DIR: str = ...
XL_MODEL_WEIGHTS_PARAM_CONFIG_PATH: str = ...
