# you should run these tests whenever you make changes to any of the train*.py files within this directory
# running these tests takes around 10 minutes on my machine
cd mixed_mm
python test_mixed_mm.py
cd ../pad_mm
python test_pad_mm.py
cd ../
