#include <torch/enum.h>

TORCH_ENUM_DEFINE(Linear)
TORCH_ENUM_DEFINE(Conv1D)
TORCH_ENUM_DEFINE(Conv2D)
TORCH_ENUM_DEFINE(Conv3D)
TORCH_ENUM_DEFINE(ConvTranspose1D)
TORCH_ENUM_DEFINE(ConvTranspose2D)
TORCH_ENUM_DEFINE(ConvTranspose3D)
TORCH_ENUM_DEFINE(Sigmoid)
TORCH_ENUM_DEFINE(Tanh)
TORCH_ENUM_DEFINE(ReLU)
TORCH_ENUM_DEFINE(GELU)
TORCH_ENUM_DEFINE(LeakyReLU)
TORCH_ENUM_DEFINE(FanIn)
TORCH_ENUM_DEFINE(FanOut)
TORCH_ENUM_DEFINE(Constant)
TORCH_ENUM_DEFINE(Reflect)
TORCH_ENUM_DEFINE(Replicate)
TORCH_ENUM_DEFINE(Circular)
TORCH_ENUM_DEFINE(Nearest)
TORCH_ENUM_DEFINE(Bilinear)
TORCH_ENUM_DEFINE(Bicubic)
TORCH_ENUM_DEFINE(Trilinear)
TORCH_ENUM_DEFINE(Area)
TORCH_ENUM_DEFINE(NearestExact)
TORCH_ENUM_DEFINE(Sum)
TORCH_ENUM_DEFINE(Mean)
TORCH_ENUM_DEFINE(Max)
TORCH_ENUM_DEFINE(None)
TORCH_ENUM_DEFINE(BatchMean)
TORCH_ENUM_DEFINE(Zeros)
TORCH_ENUM_DEFINE(Border)
TORCH_ENUM_DEFINE(Reflection)
TORCH_ENUM_DEFINE(RNN_TANH)
TORCH_ENUM_DEFINE(RNN_RELU)
TORCH_ENUM_DEFINE(LSTM)
TORCH_ENUM_DEFINE(GRU)
TORCH_ENUM_DEFINE(Valid)
TORCH_ENUM_DEFINE(Same)
